import argparse
import logging

from cenao.app import Application
from cenao.config import Config


class ApplicationRunner:
    app: Application

    def __init__(self, app: Application):
        self.app = app

    def run(self):
        parser = argparse.ArgumentParser(description='Run a cenao application')
        parser.add_argument('-c', '--config', help='Application configuration', default=None)
        parser.add_argument(
            '-e', '--env-prefix',
            help='Prefix of environment variables which would redefine config values',
            default='APP',
        )
        args = parser.parse_args()

        config = Config(args.config)
        config.process_env(args.env_prefix)

        self._init_logger(config)
        self.app.init(config)
        self.app.run()

    def _init_logger(self, config):
        config = config.get('logging', {})
        fmt = config.get('format', '%(asctime)s [%(levelname)s] [%(name)s] %(message)s')
        level = config.get('level', 'INFO')
        file = config.get('file', f'{self.app.NAME.lower()}.log')
        stdout = config.get('stdout', True)

        logging.basicConfig(
            level=level,
            filename=file,
            format=fmt
        )

        if stdout:
            handler = logging.StreamHandler()
            handler.setLevel(level)
            formatter = logging.Formatter(fmt)
            handler.setFormatter(formatter)

            logger = logging.getLogger('')
            logger.addHandler(handler)



import os
from typing import Optional, Any, Dict

import yaml
from pathlib import Path

from cenao.exceptions import ConfigNotExistsException


class Config:
    """
    Config handler and processor
    """

    def __init__(self, filename: Optional[str]):
        self.config = {}

        if filename:
            path = Path(filename)
            if not path.exists():
                raise ConfigNotExistsException(path)

            _yaml = yaml.safe_load(path.open())
            if not _yaml:
                _yaml = {}
            self.config = _yaml

    def __getitem__(self, item) -> Dict[str, Any]:
        return self.get(item)

    def get(self, item, default=None) -> Dict[str, Any]:
        """
        Returns item from config.
        :param item: Key to get from config
        :param default: Default value if key is not exists
        :return: Value of config key
        """
        return self.config.get(item, default)

    def process_env(self, env_prefix: str):
        """
        Scan environment variables and enrich config by deconstructing it's path.
        Example: `APP_WEB_PORT`'s value will be written into config by path `web.port`
        :param env_prefix: prefix to filter out variables. Default is `APP`
        """
        for e in os.environ:
            if e.startswith(env_prefix):
                entry_value = os.environ.get(e)
                entry_path = e.lower().split('_')[1:]

                path = self.config
                entry_path_len = len(entry_path)
                for i in range(entry_path_len):
                    entry = entry_path[i]
                    if entry_path_len - i != 1:
                        if entry not in path:
                            path[entry] = {}
                        path = path[entry]
                    else:
                        path[entry] = entry_value



import asyncio
import logging
import signal
import uvloop

from abc import ABC
from typing import Any, Dict, List, Optional, Type, Union

from cenao.config import Config
from cenao.exceptions import InvalidFeatureException
from cenao.view import View


class AppFeature(ABC):
    NAME: str
    CONF: Optional[str] = None
    VIEWS: List[View] = []

    app: 'Application'
    logger: logging.Logger

    def __init__(self, name: Optional[str] = None):
        if name:
            self.NAME = name.lower()

    @property
    def name(self):
        return self.NAME.lower()

    @property
    def config(self) -> Dict[str, Any]:
        return self.app.config.get(self.config_group, {})

    @property
    def config_group(self) -> str:
        if not self.CONF:
            return self.name
        else:
            return self.CONF.lower()

    def init(self, app: 'Application'):
        self.app = app
        self.logger = logging.getLogger(self.NAME)
        self.on_init()

    def on_init(self):
        pass

    async def on_startup(self):
        pass

    async def on_shutdown(self):
        pass

    async def check_health(self) -> bool:
        return True


class Features(dict):
    def __init__(self):
        super().__init__()
        self.__dict__ = self

    def __iter__(self):
        return iter(self.values())


class Application(ABC):
    NAME: str
    FEATURES: List[Union[AppFeature, Type[AppFeature]]]

    config: Config
    app_config: Dict[str, Any]
    logger: logging.Logger
    ft: Features = Features()
    loop: asyncio.AbstractEventLoop
    _task: Optional[asyncio.Task] = None

    def init_features(self):
        for feature in self.FEATURES:
            if isinstance(feature, AppFeature):
                self.logger.info('Init %s', feature.__class__.__name__)
                _feature = feature
            elif isinstance(feature(), AppFeature):
                self.logger.info('Init %s', feature.__name__)
                _feature = feature()
            else:
                raise InvalidFeatureException(feature)

            _feature.init(self)
            self.ft[_feature.name] = _feature

    def init(self, config):
        self.config = config
        self.app_config = config.get('app')
        self.logger = logging.getLogger(self.NAME + 'App')
        self.logger.info('Init application')

        if self.app_config.get('use_uvloop', True):
            self.logger.info('Installing uvloop')
            uvloop.install()

        self.loop = asyncio.get_event_loop()
        self.loop.add_signal_handler(signal.SIGINT, self.shutdown)
        self.loop.add_signal_handler(signal.SIGTERM, self.shutdown)

        self.on_init()
        self.init_features()

    def run(self):
        self.logger.info('Application started')
        features: List[AppFeature] = []
        ft: AppFeature
        for ft in self.ft:
            self.logger.info('Startup feature %s', ft.name)
            self._task = self.loop.create_task(ft.on_startup())
            try:
                self.loop.run_until_complete(self._task)
            except asyncio.CancelledError:
                pass
            except Exception as e:
                self.logger.exception('Got exception while feature startup: %r', e, exc_info=e)
                break
            features.append(ft)

        for ft in reversed(features):
            self.logger.info('Shutdown feature %s', ft.name)
            self._task = self.loop.create_task(ft.on_shutdown())
            try:
                self.loop.run_until_complete(self._task)
            except asyncio.CancelledError:
                pass
            except Exception as e:
                self.logger.exception('Got exception while feature shutdown: %r', e, exc_info=e)
                break

        self.logger.info('Application stopped')

    def shutdown(self):
        self.logger.info('Shutdown requested')
        if self._task:
            self._task.cancel()

    def on_init(self):
        pass



import re
from pathlib import Path
from setuptools import find_packages, setup


def get_version() -> str:
    file = Path("src/cenao/version.py")
    content = file.read_text(encoding="utf-8")

    m = re.search(r'^\s*__version__\s*\=\s*[\'"]([^\'""]+)[\'"]', content, re.M)
    assert m
    return m.group(1)


setup(
    name='cenao',
    version=get_version(),
    url='https://gitlab.uwtech.org/uwtech/cenao',
    license='MIT',
    author='Roman Shishkin',
    author_email='spark@uwtech.org',
    description='Python framework for fast and async applications',
    project_urls={
        'Source': 'https://gitlab.uwtech.org/uwtech/cenao',
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Framework :: AsyncIO",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP",
    ],
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    python_requires='>=3.8',
    install_requires=[
        'aiohttp==3.8.1',
        'async-timeout==4.0.2',
        'PyYAML==6.0',
        'uvloop==0.16.0',
    ],
    extras_require={
        'redis': [
            'aioredis==1.3.1',
            'aioredis-cluster==1.7.1',
        ],
    }
)



import asyncio
import logging
from typing import List

from aiohttp import web
from aiohttp.web_runner import TCPSite, AppRunner

from cenao.app import AppFeature
from cenao.view import View


class WebAppFeature(AppFeature):
    NAME = 'web'

    views: List[View] = []
    host: str
    port: int

    aiohttp_app: web.Application
    runner: AppRunner

    def on_init(self):
        self.host = self.config.get('host', '0.0.0.0')
        self.port = int(self.config.get('port', 8000))

        app = web.Application(
            loop=self.app.loop
        )
        routes_len = 0
        for ft in self.app.ft:
            for view in ft.VIEWS:
                view.init(ft)
                app.router.add_view(view.ROUTE, view)
                routes_len += 1
        self.logger.info('Registered %d routes', routes_len)

        self.runner = AppRunner(app, logger=self.logger)

    async def on_startup(self):
        self.logger.info('Starting on %s:%d', self.host, self.port)
        await self.runner.setup()
        site = TCPSite(
            self.runner,
            self.host,
            self.port,
        )
        await site.start()
        while True:
            await asyncio.sleep(3600)

    async def on_shutdown(self):
        self.logger.info('Stopping webserver')
        await self.runner.cleanup()



import async_timeout
from asyncio import Lock, Queue
from contextlib import asynccontextmanager
from typing import AsyncGenerator, AsyncContextManager, Any, Dict, List, Optional

import aioredis

from cenao.app import AppFeature


class TeeChannel:
    _tee: 'Tee'
    _queue: Queue

    def __init__(self, tee: 'Tee'):
        self._tee = tee
        self._queue = Queue()

    async def get(self) -> Any:
        if self._queue.empty() and not self._tee.lock.locked():
            async with self._tee.lock:
                await self._tee.get()

        return await self._queue.get()

    async def get_timeout(self, timeout: float):
        async with async_timeout.timeout(timeout):
            return await self.get()

    def put(self, item):
        self._queue.put_nowait(item)

    async def __anext__(self) -> Any:
        return await self.get()


class Tee:
    redis: aioredis.Redis
    channel: str
    lock: Lock = Lock()
    _channel: Optional[aioredis.Channel]
    _channels: List[TeeChannel] = []

    def __init__(self, redis: aioredis.Redis, channel: str):
        self.redis = redis
        self.channel = channel

    async def get(self):
        if not self._channel:
            raise RuntimeError('Must initialize channel first')

        item = await self._channel.get()
        for ch in self._channels:
            ch.put(item)

    async def _start(self):
        channels = await self.redis.subscribe(self.channel)
        self._channel = channels[0]

    async def _stop(self):
        try:
            await self.redis.unsubscribe(self.channel)
        except Exception:
            pass
        finally:
            self._channel = None

    @asynccontextmanager
    async def get_channel(self) -> AsyncGenerator[TeeChannel, None]:
        if self.empty():
            await self._start()

        tee_channel = TeeChannel(self)
        self._channels.append(tee_channel)

        try:
            yield tee_channel
        finally:
            self._channels.remove(tee_channel)
            if self.empty():
                await self._stop()

    def empty(self) -> bool:
        return len(self._channels) == 0


class TeeKeeper:
    redis: aioredis.Redis
    _tees: Dict[str, Tee] = {}

    def __init__(self, redis: aioredis.Redis):
        self.redis = redis

    @asynccontextmanager
    async def get_tee_channel(self, channel: str) -> AsyncGenerator[TeeChannel, None]:
        if channel not in self._tees:
            self._tees[channel] = Tee(self.redis, channel)
        tee = self._tees[channel]

        try:
            async with tee.get_channel() as tee_channel:
                yield tee_channel
        finally:
            if tee.empty():
                del self._tees[channel]


class RedisAppFeature(AppFeature):
    """
    Cenao feature for Redis support.
    By default tries connect to `redis://172.17.0.1`, which can be changed via `nodes` config parameter.
    """
    NAME = 'redis'

    _client: aioredis.Redis
    _tee_keeper: Optional[TeeKeeper] = None

    async def on_startup(self):
        cluster = self.config.get('cluster', False)
        nodes = self.config.get('nodes', 'redis://172.17.0.1').split(',')

        if cluster:
            raise NotImplementedError('aioredis-cluster support not implemented yet')
        else:
            self._client = await aioredis.create_redis_pool(nodes[0])

        await self._client.ping()

    async def on_shutdown(self):
        if self._client:
            self._client.close()
            await self._client.wait_closed()

    async def check_health(self) -> bool:
        await self._client.ping()
        return True

    @property
    def client(self):
        """
        Redis instance getter
        :return: A Redis client instance
        """
        return self._client

    def subscribe(self, channel: str) -> AsyncContextManager[TeeChannel]:
        """
        Subscribe (and reuse existing connection) to Redis pubsub channel
        :return: A Tee channel handler
        """
        if not self._client:
            raise RuntimeError('Redis connection must be established first')

        if not self._tee_keeper:
            self._tee_keeper = TeeKeeper(self.client)

        return self._tee_keeper.get_tee_channel(channel)
